
## Weiler-Atherton Clipping Applet

## Author: Christopher Andrews
## Date: May 2008

from java.awt import Color
from Graphics import CVPolygon
import Utilities


class ClipEngine:
    def __init__(self, controller):
        self.controller = controller
        
    def insertIntersection(self,subject,clipping, A,B,C,D):
        """
        This is looking for the intersection point between the line segments
        AB and CD. We are using the parametric equation A-bt = C - du, where
        b is the vector (B-A) and d is the vector (D-C). We want to solve for
        t and u - if they are both between 0 and 1, we have a valid
        intersection point.

        For insertion, we assume that AB comes from the subject polygon and
        that CD comes from the clipping region. This means that the
        intersection point can be inserted directly between CD but that we need
        to use the t value to find a valid position between A and B.
        """
        self.controller.actionStack.append((self.controller.FIND_INTER,'Comparing the lines defined by the segments %s%s and %s%s' % (A.name, B.name, C.name, D.name), ('showLines', A, B, C, D)))
        intersection = Utilities.intersection(A,B,C,D)
        if intersection is None:
            # the lines are parallel
            self.controller.actionStack.append((self.controller.FIND_INTER,'%s%s and %s%s are parallel, so they are ignored' % (A.name, B.name, C.name, D.name), ('showLines', A, B, C, D)))
            return

        (t,u,inbound) = intersection
        b = B - A
        intersection = A+b*t

        self.controller.actionStack.append((self.controller.FIND_INTER,'Calculating the intersection between %s%s and %s%s'% (A.name, B.name, C.name, D.name) , ('showLines', A, B, C, D), ('showPoint',intersection)))

        if u is None:
            # we know that the point isn't within t
            self.controller.actionStack.append((self.controller.FIND_INTER,'The intersection point is not contained within %s%s, so it is thrown out' % (A.name, B.name), ('showSegments', A, B, C, D), ('showPoint',intersection)))
            return

        # the point is on line segment AB
        if 0 < u < 1:
            # we have a valid intersection point
            if t == 0:
                newPoint = A
                newPoint.setT(t, subject.name)
            elif t==1:
                newPoint = B
                newPoint.setT(t, subject.name)
            else:
                newPoint = intersection
                newPoint.setT(t, subject.name)
                newPoint.inbound = inbound

                # insert it into AB
                prev = A
                next = A.next[subject.name]
                while next is not None and next.t[subject.name] < t:
                    prev = next
                    next = next.next[subject.name]

                prev.setNext(newPoint, subject.name)
                newPoint.setNext(next, subject.name)

            newPoint.setT(u, clipping.name)
            #insert into CD
            prev = C
            next = C.next[clipping.name]
            while next is not None and next.t[clipping.name] < u:
                prev = next
                next = next.next[clipping.name]

            prev.setNext(newPoint, clipping.name)
            newPoint.setNext(next, clipping.name)

            newPoint.name = str(self.currentIntersection)
            self.currentIntersection += 1
            self.controller.actionStack.append((self.controller.FIND_INTER,'The intersection point is contained within both line segments, so it is inserted into both polygons', ('showSegments', A, B, C, D), ('showPoint',newPoint,True)))
        elif (u == 0 or u == 1) and (0<t<1):
            # we can ignore them if the point it an endpoint in both
            # polygons - any following we do will just wrap around
            if u == 0:
                newPoint = C
            else:
                newPoint = D

            newPoint.setT(t, subject.name)
            newPoint.setT(u, clipping.name)
            newPoint.inbound = inbound

            # insert it into AB
            prev = A
            next = A.next[subject.name]
            while next is not None and next.t[subject.name] < t:
                prev = next
                next = next.next[subject.name]

            prev.setNext(newPoint, subject.name)
            newPoint.setNext(next, subject.name)
            newPoint.name = str(self.currentIntersection)
            self.currentIntersection += 1
            self.controller.actionStack.append((self.controller.FIND_INTER,'The intersection point is contained within both line segments, so it is inserted into both polygons', ('showSegments', A, B, C, D), ('showPoint',newPoint, True)))
        else:
            self.controller.actionStack.append((self.controller.FIND_INTER,'The intersection point is not contained within %s%s, so it is thrown out'%(C.name, D.name), ('showSegments', A, B, C, D), ('showPoint',intersection)))
 
            

 
    def findIntersections(self, subject, clipping):
        #import pdb
        #pdb.set_trace()
        self.currentIntersection = 1
        A = subject.pointList
        while A is not None:
            B = A.next[subject.name]
            tmp = B
            if B is None:
                B = subject.pointList
            C =  clipping.pointList
            while C is not None:
                D = C.next[clipping.name]
                while D is not None and D.isIntersection(clipping.name):
                    # skip past intersections
                    D = D.next[clipping.name]
                if D is not None:
                    self.insertIntersection(subject,clipping,A,B,C,D)
                else:
                    self.insertIntersection(subject,clipping,A,B,C,clipping.pointList) # loopback on clipping
                C = D
            A = tmp

    def findEnteringIntersection(self, plist, name, clipRegion):
        if plist is None:
            return None
        current = plist
      
        while current is not None and not (current.inbound and not current.visited):
            if current.inbound:
                self.controller.actionStack.append((self.controller.CLIP,"Point %s has already been visited, so advance to the next point" % current.name, ('showPoint', current)))
            else:
                self.controller.actionStack.append((self.controller.CLIP,"Point %s is not an inbound intersection, so advance to the next point" % current.name, ('showPoint', current)))
            current = current.next[name]
        return current   
    
    def clip(self, p0, p1):
        """
      The main clipping function        
        """
      
        self.controller.actionStack.append((self.controller.FIND_INTER,'Step one is to identify all of the intersection points between the two polygons',))
        # find intersections
        self.findIntersections(p1,p0)
        self.controller.actionStack.append((self.controller.FIND_INTER,'The intersections have now all been identified', ('saveIntersections',)))
        polygons = []
        names = (p0.name, p1.name)
        # get next entering intersection
        entering = self.findEnteringIntersection(p1.pointList, p1.name, p0)
        while entering is not None:
            self.controller.actionStack.append((self.controller.CLIP,"Point %s is an inbound intersection" % entering.name, ('showPoint', entering)))
            newPoly = CVPolygon(Color.green)
            current = entering
            currentPoly = 1
            while not current.visited:
                self.controller.actionStack.append((self.controller.CLIP,"Save %s and mark it as visited" % current.name, ('showPoint', current, True)))
                newPoly.addPoint(current.x, current.y, current.name)
         
                current.visited = True

                if current.isIntersection(names[currentPoly]) and current != entering:
                    currentPoly = (currentPoly + 1) % 2
                    self.controller.actionStack.append((self.controller.CLIP,"Point %s is an intersection, so switch polygon lists, now following polygon %s" % (current.name,names[currentPoly]),('switchPolygon',names[currentPoly])))
              
                current = current.next[names[currentPoly]]
                if current is None:
                    # time to wrap
                    if currentPoly == 0:
                        current = p0.pointList
                    else:
                        current = p1.pointList
            self.controller.actionStack.append((self.controller.CLIP,"Point %s has already been visited, save this list of points as a new polygon" % (current.name), ('showPoint', current, True), ('savePolygon', newPoly)))
            self.controller.actionStack.append((self.controller.CLIP,"Switch back to SUBJECT polygon to look for next intersection",('showPoint', current), ('switchPolygon','SUBJECT') ))
            polygons.append(newPoly)
            
            entering = self.findEnteringIntersection(current.next[p1.name], p1.name, p0)

        if len(polygons) == 0:
            # we  found no polygons - possibly one is contained in the other
            # first check to see if our clipping region contains the other one
            # theoretically, any point will do, except that all of the end
            # points might intersect the boundary of the shape - and thus fail
            # the containment test. So, we pick a point that is halfway along
            # the first edge and one unit vector in the direction opposite the
            # perp vector
            self.controller.actionStack.append((self.controller.CLIP,"No inbound intersections found - checking if one polygon contains the other",))
            A = p1.pointList
            B = p1.pointList.next[p1.name]
            b = B - A
            b_perp = b.perp()
            b_perp.normalize()
            x = .5 * (A.x + B.x) - b_perp.x
            y = .5 * (A.y + B.y) - b_perp.y
           
            if p0.contains(x,y):
                # make a copy of poly1
                newPoly = CVPolygon(c=Color.green, name='SUBJECT')
                points = p1.pointList
                for p in points:
                    newPoly.addPoint(p.x,p.y, p.name)
                self.controller.actionStack.append((self.controller.CLIP,"The SUBJECT polygon is contained completely inside the CLIP polygon, so it is returned", ('savePolygon', newPoly)))
                polygons.append(newPoly)
            # now check to see if the clipping region is enclosed by the other
            else:
                A = p0.pointList
                B = p0.pointList.next[p0.name]
                b = B - A
                b_perp = b.perp()
                b_perp.normalize()
                x = .5 * (A.x + B.x) - b_perp.x
                y = .5 * (A.y + B.y) - b_perp.y
               
                if p1.contains(x,y):
                    # make a copy of poly0
                    newPoly = CVPolygon(c=Color.green, name='CLIP')
                    points = p0.pointList
                    for p in points:
                        newPoly.addPoint(p.x,p.y, p.name)
                    self.controller.actionStack.append((self.controller.CLIP,"The CLIP polygon is contained completely inside the SUBJECT polygon, so it is returned", ('savePolygon', newPoly)))
                    polygons.append(newPoly)
                else:
                    self.controller.actionStack.append((self.controller.CLIP,"The polygons are disjoint - no polygons returned",))
        self.controller.actionStack.append((self.controller.CLIP,"There are no more unvisited inbound intersections", ("finish",)))
        return polygons
