
## Weiler-Atherton Clipping Applet

## Author: Christopher Andrews
## Date: May 2008


## To build applet: jythonc --deep --core --jar clipping.jar *.py
## To include resources: jar uf clipping.jar help


from javax.swing import JApplet, UIManager
from javax.swing import JFrame, JPanel, BorderFactory, JButton, JScrollPane, JTextArea, JOptionPane
from javax.swing import JTree, JTabbedPane, BoxLayout,Box
from javax.swing.tree import DefaultMutableTreeNode,DefaultTreeCellRenderer

from java.awt import Color, Dimension
from java.awt import  CardLayout
from java.awt.event import KeyEvent

from java.lang import Short

from DrawingPanel import DrawingPanel
from PointListPanel import PointListPanel
from ConsoleView import ConsoleView
from HelpView import HelpView
from Graphics import CVPolygon, Vector, PointNode
from ClipEngine import ClipEngine



class ClippingVisualizer(JApplet):
    """
    This class will serve as the main controller - keeping track of the GUI
    items that we need to keep track of and coordinating communication between
    them.
    """
    ENTER_CLIP = '0.a'
    ENTER_SUBJ = '0.b'
    FIND_INTER = '1'
    CLIP = '2'
    messages = {ENTER_CLIP: 'Entering CLIP polygon',
                ENTER_SUBJ: 'Entering SUBJECT polygon',
                FIND_INTER: 'Finding intersections',
                CLIP:       'Clipping'}
    
    def init(self):
        """
        This function is where we create all of the GUI widgets, build the
        layout and connect everything together
        """
        UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName())
        _frame = JFrame("Weiler-Atherton Clipping Demo", keyPressed = self.keyEvent, keyReleased = self.keyEvent)
        
        self.canvas = DrawingPanel(self)
      

        self.pointListView = PointListPanel(self)
        _plScroller = JScrollPane(self.pointListView)
        _plScroller.setPreferredSize(Dimension(500,125))
        _plScroller.setMaximumSize(Dimension(Short.MAX_VALUE, 125))
        _plScroller.setMinimumSize(Dimension(Short.MIN_VALUE, 125))

        self.controlPanel = JPanel(CardLayout())
        self.controlPanel.setBorder(BorderFactory.createTitledBorder("Controls"))
        self.controlPanel.setMaximumSize(Dimension(300,100))
        self.controlPanel.setMinimumSize(Dimension(300,100))
        self.polygonButton = JButton("Polygon complete")
        self.polygonButton.setFocusable(False)
        self.polygonButton.actionPerformed = self.polygonButtonClicked
     
        self.stepButton = JButton("Forward")
        self.stepButton.setFocusable(False)
        self.stepButton.actionPerformed = self.stepButtonClicked
        self.backButton = JButton("Back")
        self.backButton.setFocusable(False)
        self.backButton.actionPerformed = self.backButtonClicked
     
        self.resetButton = JButton("Reset")
        self.setFocusable(False)
        self.resetButton.actionPerformed = self.resetButtonClicked
        self.reset2Button = JButton("Reset")
        self.reset2Button.setFocusable(False)
        self.reset2Button.actionPerformed = self.resetButtonClicked

        _polyCard = JPanel()
        _polyCard.add(self.polygonButton)
        _polyCard.add(self.resetButton)
        self.controlPanel.add(_polyCard, "POLYGON MODE")
        _clipCard = JPanel()
        _clipCard.add(self.backButton)
        _clipCard.add(self.stepButton)
        _clipCard.add(self.reset2Button)
        self.controlPanel.add(_clipCard, "CLIP MODE")

        _tabbedViews = JTabbedPane()
        _tabbedViews.setFocusable(False)
        _tabbedViews.setPreferredSize(Dimension(300,500))
        _tabbedViews.setMaximumSize(Dimension(300, Short.MAX_VALUE))

        self.console = ConsoleView()
        self.console.setFocusable(False)
    
        self.console.addLink('simple', 'simple polygon')
        self.console.addLink('inbound', 'inbound intersection')
        _tScroller = JScrollPane(self.console)
        _tScroller.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS)
        
        
        _tabbedViews.addTab("Console", None, _tScroller, "Action log and explanitory text")


        self.help = HelpView()
        self.help.setFocusable(False)
    
        self.console.hyperlinkUpdate = self.help.hyperlinkEvent
      
        _tScroller = JScrollPane(self.help)
        _tScroller.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS)
        
        
        _tabbedViews.addTab("Help", None, _tScroller, "Documentation")
        
        _mainPanel = JPanel()
        _mainPanel.setLayout(BoxLayout(_mainPanel, BoxLayout.LINE_AXIS))

        _leftPanel = JPanel()
        _leftPanel.setLayout(BoxLayout(_leftPanel, BoxLayout.PAGE_AXIS))
        _leftPanel.add(Box.createRigidArea(Dimension(0,10)))
        _leftPanel.add(self.canvas)
        _leftPanel.add(Box.createRigidArea(Dimension(0,10)))
        _leftPanel.add(_plScroller)   
        _leftPanel.add(Box.createRigidArea(Dimension(0,10)))

    
        _rightPanel = JPanel()
        _rightPanel.setLayout(BoxLayout(_rightPanel, BoxLayout.PAGE_AXIS))
        _rightPanel.add(Box.createRigidArea(Dimension(0,10)))
        _rightPanel.add(self.controlPanel)
        _rightPanel.add(Box.createRigidArea(Dimension(0,10)))
        _rightPanel.add(_tabbedViews)
        _rightPanel.add(Box.createRigidArea(Dimension(0,10)))

        _mainPanel.add(Box.createRigidArea(Dimension(10,0)))
        _mainPanel.add(_leftPanel)
        _mainPanel.add(Box.createRigidArea(Dimension(10,0)))
        _mainPanel.add(_rightPanel)
        _mainPanel.add(Box.createRigidArea(Dimension(10,0)))
        _frame.add(_mainPanel)
        _frame.pack()
        self.reset()
        _frame.requestFocusInWindow()
        _frame.setVisible(True)
       

    def reset(self):
        self.mode = self.ENTER_CLIP # we start ready to define the clipping region
        layout = self.controlPanel.getLayout()
        layout.first(self.controlPanel)
       
        self.consoleReset()
        self.canvas.reset()
        self.pointListView.reset()
        self.actionStack = []
        self.currentAction = 0
        self.currentPhase = None
        self.numPoints = 0
        self.resetButton.setEnabled(False)
        self.backButton.setEnabled(False)
        self.polygonButton.setEnabled(False)
        self.stepButton.setEnabled(False)
        self.polygons = {'CLIP': CVPolygon(Color.lightGray, 1,'CLIP'), 'SUBJECT': CVPolygon(Color.blue, .3,'SUBJECT')}

    def consoleReset(self):

        self.console.initText('Use the mouse to define a simple polygon to be used as the clipping region - when the polygon is complete, click the "Polygon complete" button above\n')
  


    def polygonButtonClicked(self, event):
        if self.mode == self.ENTER_CLIP:
            
            self.polygonButton.setEnabled(False)
            self.numPoints = 0
            if not self.polygons['CLIP'].validate('abcdefghijklmnopqrstuvwxyz'):
                JOptionPane.showMessageDialog(self, 'The Weiler-Atherton algorithm only works with simple polygons.\nThe polygon you entered is not simple - two of the edges intersect.\nPlease enter a new polygon.', 'Bad Input', JOptionPane.ERROR_MESSAGE)
                self.polygons['CLIP'] = CVPolygon(Color.lightGray, 1,'CLIP')
            else:
                self.mode = self.ENTER_SUBJ
                self.console.initText('Now, specify a subject polygon to be clipped by the clipping region - again, use the "Polygon complete" button to indicate that the polygon is finished')
                      
        elif self.mode == self.ENTER_SUBJ:
            self.polygonButton.setEnabled(False)
            self.numPoints = 0
            if not self.polygons['SUBJECT'].validate('ABCDEFGHIJKLMNOPQRSTUVWXYZ'):
                JOptionPane.showMessageDialog(self, 'The Weiler-Atherton algorithm only works with simple polygons.\nThe polygon you entered is not simple - two of the edges intersect.\nPlease enter a new polygon.', 'Bad Input', JOptionPane.ERROR_MESSAGE)
                self.polygons['SUBJECT'] = CVPolygon(Color.blue, .3,'SUBJECT')
            else:
                self.mode = self.FIND_INTER
                self.stepButton.setEnabled(True)
                layout = self.controlPanel.getLayout()
                layout.last(self.controlPanel)

                self.console.initText('The "Forward" and "Back" buttons can now be used to walk through the algorithm. If you hold the Shift key when clicking a button, you can jump immediatly to the end or the beginning of the current phase of the algorithm')
                clippingEngine = ClipEngine(self)
                clippingEngine.clip(self.polygons['CLIP'], self.polygons['SUBJECT'])
        self.canvas.repaint()
        self.pointListView.repaint()    


    def saveIntersections(self, args):
        self.console.consoleWrite("End of intersection location phase", step=False)
                


    def finish(self, args):
        self.stepButton.setEnabled(False)
        self.console.consoleWrite("Algorithm complete", False)
        

            
    def applyAction(self, action):
        mode = action[0]
        text = action[1]
        actions = action[2:]

        self.console.consoleWrite(text)
        
        for action in actions:
            try:
                method = getattr(self,action[0])
                args = action[1:]
                if callable(method):
                    method(args)
                else:
                    self.controller.console.consoleWrite("Error: Attribute %s is not a method" % action[0])

            except AttributeError:
                # this action isn't meant for this object so ignore it
                pass
     
        self.mode = mode
        self.canvas.applyActions(actions)
        self.pointListView.applyActions(actions)
     


    def keyEvent(self, action):
        if action.isShiftDown():
            self.stepButton.setText("Phase end")
            self.backButton.setText("Phase start")
        else:
            self.stepButton.setText("Forward")
            self.backButton.setText("Back")

        

    def stepButtonClicked(self, event):
        self.backButton.setEnabled(True)
        try:
            if event.getModifiers() & event.SHIFT_MASK:
                # jump forward to next phase

                while self.actionStack[self.currentAction][0] == self.mode:
                    self.applyAction(self.actionStack[self.currentAction])
                    self.currentAction += 1
               
                self.mode = self.actionStack[self.currentAction][0]
            else:
                # just move forward a single step
                self.applyAction(self.actionStack[self.currentAction])
                self.currentAction += 1
        except IndexError:
            # out of actions
            self.stepButton.setEnabled(False)
           
        self.canvas.repaint()
        self.pointListView.repaint()

    def resetButtonClicked(self, event):
        self.reset()

    def backButtonClicked(self, event):
        self.stepButton.setEnabled(True)
        # reset the views
        self.canvas.resetActions()
        self.pointListView.resetActions()
        for polygon in self.polygons.values():
            polygon.hideIntersections()
        self.consoleReset()

        if event.getModifiers() & event.SHIFT_MASK:
            # jump back to the start of the step
            self.currentAction = 0
            currentMode = self.mode
            
            while currentMode != self.actionStack[self.currentAction][0]:
                self.applyAction(self.actionStack[self.currentAction])
                self.currentAction += 1
        else:
            # just move back one step
            self.currentAction -= 1
            i = 0
            while i < self.currentAction:
                self.applyAction(self.actionStack[i])
                i+=1

        if self.currentAction <= 0:
            self.backButton.setEnabled(False)
        self.canvas.repaint()
        self.pointListView.repaint()

        
    def canvasClick(self):
        self.numPoints += 1
        
        if self.numPoints == 1:
            self.resetButton.setEnabled(True)
        elif self.numPoints == 3:
            # user has entered at least three points for a polygon
            self.polygonButton.setEnabled(True)
        

if __name__ == '__main__':
    import pawt
    pawt.test(ClippingVisualizer())
