## Weiler-Atherton Clipping Applet

## Author: Christopher Andrews
## Date: May 2008


from javax.swing import JEditorPane
from javax.swing.text.html import HTMLEditorKit
from javax.swing.event import HyperlinkEvent
from javax.swing import JOptionPane




class ConsoleView(JEditorPane):
    def __init__(self):
        self.links = {}
        self.descriptions = {}
        self.setEditable(False)
        editorKit = HTMLEditorKit()
        style = editorKit.getStyleSheet()
        
        style.addRule("body{font-size:small;font-family:sans-serif;}span.info{font-weight:bold;}div.break{text-align:center; font-weight:bold; background-color:#3366CC; color:white;margin-top:10px;margin-bottom:10px;padding:5px;")
        self.setEditorKit(editorKit)
        self.initText('')
 
    def consoleWrite(self, text, step=True):
        """
        This is the general purpose function for adding text to the console.
        The step argument is used to treat actual steps of the algorithm
        differently from other informative data that gets printed to the
        console. At the moment, the only difference is to add a - at the start
        of steps, but something more interesting could be done here.
        """
        text = self.makeLinks(text)
        if step:
            self.txt += '<br>%s<br>' % text
        else:

            self.txt += '<div class="break">%s</div>'% text
        self.setText(self.txt)

    def initText(self, text):
        """
        This function wipes the contents of the console and replaces it with
        the contents of text after performing a makeLinks on it.
        """
        self.txt = '<span class="info">%s</span><br>' % self.makeLinks(text)
        self.setText(self.txt)

  
    def addLink(self, link, text):
        """
        This function adds a phrase to use as a link to the console. The text
        is the phrase that needs to be explained and the link is a single word
        term that is used to make a hyperlink.
        """
        self.links[text] = link
       

    def makeLinks(self, text):
        """
        This function takes a piece of text and searches through it for any
        occurrances of the set key phrases we are interested in. The phrase
        is replaced with a link of the same text, with a URL that can be used
        as a key inot the collection of descriptions.
        """
      
        snippets = self.links.keys()
        for snippet in snippets:
            l = text.split(snippet)
            link = '<a href="%s">%s</a>' % (self.links[snippet], snippet)
            text = link.join(l)
        return text
