## Weiler-Atherton Clipping Applet

## Author: Christopher Andrews
## Date: May 2008


def intersection(A, B, C, D):
    """
    This calculates the intersection point between two lines, AB and CD.
    The function is somewhat specialized for the purposes of the clipping
    visualization. The return value is None if the lines are parallel or a
    tuple of the form (t, u, inbound). inbound is a boolean value if
    we know that CD is a line that belongs to a polygon specified in CCW
    direction. The further optimization is that u and inbound are set to
    None if t is not within 0 and 1. All of the uses of this function
    in this application are only interested in line segments and this saves
    a couple of extra calculations.
    """
    b = B - A
    d_perp = (D - C).perp()
    denom = d_perp.dot(b)
    if denom != 0:
        c = C - A
        numer =  d_perp.dot(c)
        t = float(numer) / float(denom)
        if 0 <= t <= 1:
            b_perp = b.perp()
            numer = b_perp.dot(c)
            u = float(numer) / float(denom)
            return (t, u, denom < 0)
        else:
            # the intersetion isn't within segment AB, so bail out
            return (t, None, None)
    else:
        # the lines are parallel
        return None


